==> gau-0.2b.tar.gz: gau-0.2/chartable.c <==

/* this file is a part of gau software, (C) Hwang chi-duck 1997               */

#include <glib.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtksignal.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <string.h>

#include "chartable.h"

#define PAD 6

static void char_table_class_init    (CharTableClass *klass);
static void char_table_init          (CharTable      *darea);
static void char_table_realize       (GtkWidget           *widget);
static void char_table_destroy       (GtkObject *object);
static void char_table_size_allocate (GtkWidget           *widget,
					    GtkAllocation       *allocation);
static void char_table_size_request (GtkWidget           *widget,
				      GtkRequisition *requisition);
static gint char_table_expose          (GtkWidget      *widget,
				      GdkEventExpose *event);
static gint char_table_button_press          (GtkWidget      *widget,
				      GdkEventButton *event);
static gint char_table_key_press	(GtkWidget      *widget,
				      GdkEventKey *event);
static void char_table_item_select(CharTable *table);
static void char_table_item_unselect(CharTable *table);
static void char_table_item_draw(CharTable *table, int x, int y, int selected);

static GtkWidgetClass *parent_class = NULL;

enum {
	SELECTED,
	LAST_SIGNAL
};

static gint char_table_signals[LAST_SIGNAL] = { 0 };

typedef void (*CharTableSignal1) (GtkObject *object,
	gpointer arg1,
	gpointer data);

static void char_table_marshal_signal_1 (GtkObject *object,
	GtkSignalFunc func,
	gpointer func_data,
	GtkArg	*arg);

guint
char_table_get_type ()
{
  static guint char_table_type = 0;

  if (!char_table_type)
    {
      GtkTypeInfo char_table_info =
      {
	"CharTable",
	sizeof (CharTable),
	sizeof (CharTableClass),
	(GtkClassInitFunc) char_table_class_init,
	(GtkObjectInitFunc) char_table_init,
	(GtkArgFunc) NULL,
      };

      char_table_type = gtk_type_unique (gtk_widget_get_type (), &char_table_info);
    }
  // g_print("%s in %s\n", __FUNCTION__, __FILE__);

  return char_table_type;
}

static void
char_table_class_init (CharTableClass *class)
{
	GtkWidgetClass *widget_class;
	GtkObjectClass *object_class;

	object_class = (GtkObjectClass*) class;
	widget_class = (GtkWidgetClass*) class;
	//g_print("%s in %s\n", __FUNCTION__, __FILE__);

	char_table_signals[SELECTED] = gtk_signal_new("selected",
		GTK_RUN_LAST,
		object_class->type,
		GTK_SIGNAL_OFFSET (CharTableClass, selected),
		char_table_marshal_signal_1, GTK_TYPE_NONE, 1,
		GTK_TYPE_POINTER);
	gtk_object_class_add_signals (object_class, char_table_signals, LAST_SIGNAL);
	parent_class = gtk_type_class (gtk_widget_get_type ());
	object_class->destroy = char_table_destroy;
	widget_class->realize = char_table_realize;
	widget_class->size_allocate = char_table_size_allocate;
	widget_class->size_request = char_table_size_request;
	widget_class->expose_event = char_table_expose;
	widget_class->button_press_event = char_table_button_press;
	widget_class->key_press_event = char_table_key_press;
}

static void
char_table_init (CharTable *chartable)
{
	GTK_WIDGET_SET_FLAGS(chartable, GTK_CAN_FOCUS);
	chartable->current_x = 0;
	chartable->current_y = 0;
	//g_print("%s in %s\n", __FUNCTION__, __FILE__);
}


GtkWidget *
char_table_new(char **buf, int col, int row)
{
	CharTable *table;
	GdkFont *font;
	table = gtk_type_new(char_table_get_type());
	table->col = col;
	table->row = row;
	table->buf = buf;
	font = GTK_WIDGET(table)->style->font;
	table->per_width = gdk_string_width(font, "") + PAD * 2;
	table->per_height = font->ascent + font->descent + PAD * 2;
	return GTK_WIDGET(table);
}

char *
char_table_get_char(CharTable *table)
{
	if(table->current_x < 0) return NULL;
	return &(table->buf[table->current_x][table->current_y*2]);
}

static void
char_table_realize (GtkWidget *widget)
{
	CharTable *table;
	GdkWindowAttr attributes;
	gint attributes_mask;
	//g_print("%s in %s\n", __FUNCTION__, __FILE__);
	g_return_if_fail (widget != NULL);

	table = CHAR_TABLE(widget);
	GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);

	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.visual = gtk_widget_get_visual (widget);
	attributes.colormap = gtk_widget_get_colormap (widget);
	attributes.event_mask = gtk_widget_get_events (widget);
	attributes.event_mask |= (GDK_EXPOSURE_MASK |
		GDK_BUTTON_PRESS_MASK |
		GDK_KEY_PRESS_MASK);

	attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

	widget->window = gdk_window_new (widget->parent->window, &attributes, attributes_mask);
	gdk_window_set_user_data (widget->window, table);
	widget->style = gtk_style_attach (widget->style, widget->window);
	gdk_window_set_background(widget->window, &widget->style->white);
}

static void
char_table_size_allocate (GtkWidget     *widget,
				GtkAllocation *allocation)
{
  //g_print("%s in %s\n", __FUNCTION__, __FILE__);
	g_return_if_fail (widget != NULL);
	widget->allocation = *allocation;

	if (GTK_WIDGET_REALIZED (widget)) {
		gdk_window_move_resize (widget->window,
			allocation->x, allocation->y,
			allocation->width, allocation->height);
    }
}

static gint
char_table_expose(GtkWidget *w, GdkEventExpose *ev)
{
	int x1, x2;
	int y1, y2;
	int i, j;
	CharTable *table = CHAR_TABLE(w);
	x1 = ev->area.x / table->per_width;
	x2 = (ev->area.x + ev->area.width) / table->per_width;
	y1 = ev->area.y / table->per_height;
	y2 = (ev->area.y + ev->area.height) / table->per_height;
	if(y2 >= table->row) y2 = table->row - 1;
	if(x2 >= table->col) x2 = table->col - 1;
	for(i=x1; i<= x2; i++) for(j=y1;j<=y2;j++) {
		if(i == table->current_x && j == table->current_y) {
			char_table_item_select(table);
		} else
			char_table_item_draw(table, i, j, FALSE);
	}
	return FALSE;
}

static gint
char_table_button_press(GtkWidget *w, GdkEventButton *ev)
{
	CharTable *table = CHAR_TABLE(w);
	int row, col;

	if (!GTK_WIDGET_HAS_FOCUS (w))
		gtk_widget_grab_focus (w);
	col = ev->x / (table->per_width);
	row = ev->y / (table->per_height);
	if(row < 0 || row >= table->row) return FALSE;
	if(col < 0 || col >= table->col) return FALSE;
	if(ev->type == GDK_2BUTTON_PRESS) {
		gtk_signal_emit(GTK_OBJECT(w), char_table_signals[SELECTED], 
			&table->buf[table->current_x][2*table->current_y]);
		return TRUE;
	}
	if(table->current_x != col || table->current_y != row) {
		char_table_item_unselect(table);
		table->current_x = col;
		table->current_y = row;
		char_table_item_select(table);
	}
	if(ev->type == GDK_2BUTTON_PRESS || ev->button == 2) {
		gtk_signal_emit(GTK_OBJECT(w), char_table_signals[SELECTED], 
			&table->buf[table->current_x][2*table->current_y]);
		return TRUE;
	}
	return TRUE;
}

static gint
char_table_key_press(GtkWidget *w, GdkEventKey *ev)
{
	CharTable *table = CHAR_TABLE(w);
	if(ev->keyval == GDK_Left) {
		char_table_item_unselect(table);
		if(table->current_x == 0) {
			table->current_x = table->col -1;
			if(table->current_y == 0) {
				table->current_y = table->row -1;
			} else 
				table->current_y--;
		} else
			table->current_x--;
	} else if(ev->keyval == GDK_Right) {
		char_table_item_unselect(table);
		if(table->current_x == table->col -1) {
			table->current_x = 0;
			if(table->current_y == (table->row -1)) {
				table->current_y = 0;
			} else
				table->current_y++;
		} else
			table->current_x++;
	} else if(ev->keyval == GDK_Up) {
		char_table_item_unselect(table);
		if(table->current_y == 0) {
			table->current_y = table->row -1;
		} else 
			table->current_y--;
	} else if(ev->keyval == GDK_Down) {
		char_table_item_unselect(table);
		if(table->current_y == (table->row -1)) {
			table->current_y = 0;
		} else
			table->current_y++;
	} else if(ev->keyval == GDK_space || ev->keyval == GDK_Return) {
		gtk_signal_emit(GTK_OBJECT(w), char_table_signals[SELECTED], 
			&table->buf[table->current_x][2*table->current_y]);
	} else
		return FALSE;
	char_table_item_select(table);
	return TRUE;
}

static void
char_table_destroy(GtkObject *object)
{
	//CharTable *chartable = CHAR_TABLE(object);

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
char_table_size_request(GtkWidget *widget, GtkRequisition *requisition)
{
	CharTable *chartable = CHAR_TABLE(widget);
  //g_print("%s in %s\n", __FUNCTION__, __FILE__);
	g_return_if_fail (widget != NULL);
	requisition->width = chartable->col * chartable->per_width;
	requisition->height = chartable->row * chartable->per_height;
}

static void
char_table_item_select(CharTable *table)
{
	GtkWidget *w = GTK_WIDGET(table);
	if(table->current_x < 0 || table->current_y < 0) return;
	if(table->current_x >= table->col || table->current_y >= table->row) return;
	gdk_draw_rectangle(w->window, w->style->bg_gc[GTK_STATE_SELECTED],
		TRUE, 
		table->per_width * table->current_x + 1, // x
		table->per_height * table->current_y + 1, // y
		table->per_width - 2,
		table->per_height - 2);
	char_table_item_draw(table, table->current_x, table->current_y, TRUE);
}

static void
char_table_item_unselect(CharTable *table)
{
	GtkWidget *w = GTK_WIDGET(table);
	if(table->current_x < 0 || table->current_y < 0) return;
	if(table->current_x >= table->col || table->current_y >= table->row) return;
	gdk_window_clear_area(w->window, 
		table->per_width * table->current_x,
		table->per_height * table->current_y,
		table->per_width, table->per_height);
	char_table_item_draw(table, table->current_x, table->current_y, FALSE);
}

static void
char_table_item_draw(CharTable *table, int x, int y, int selected)
{
	GtkWidget *w = GTK_WIDGET(table);
	gdk_draw_rectangle(w->window, w->style->black_gc,
		FALSE, 
		table->per_width * x, // x
		table->per_height * y, // y
		table->per_width - 2,
		table->per_height - 2);
	if(selected) {
		gdk_draw_text(w->window, w->style->font, w->style->fg_gc[GTK_STATE_SELECTED],
			table->per_width * x + PAD,
			table->per_height * y + PAD + w->style->font->ascent,
			&table->buf[x][2*y], 2);
	} else {
		gdk_draw_text(w->window, w->style->font, w->style->fg_gc[GTK_STATE_NORMAL],
			table->per_width * x + PAD,
			table->per_height * y + PAD + w->style->font->ascent,
			&table->buf[x][2*y], 2);
	}
}

static void
char_table_marshal_signal_1 (GtkObject *object,
	GtkSignalFunc func,
	gpointer func_data,
	GtkArg *args)
{
	CharTableSignal1 rfunc;
	rfunc = (CharTableSignal1)func;
	(*rfunc) (object, GTK_VALUE_POINTER(args[0]), func_data);
}

==> gau-0.3.1.tar.gz: gau-0.3.1/chartable.c <==

/* this file is a part of gau software, (C) Hwang chi-deok 1997, 1998         */

#include <gtk/gtkwidget.h>
#include <gtk/gtksignal.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <string.h>

#include "chartable.h"

#define PAD 8

static void char_table_class_init    (CharTableClass *klass);
static void char_table_init          (CharTable      *darea);
static void char_table_realize       (GtkWidget           *widget);
static void char_table_destroy       (GtkObject *object);
static void char_table_size_allocate (GtkWidget           *widget,
					    GtkAllocation       *allocation);
static void char_table_size_request (GtkWidget           *widget,
				      GtkRequisition *requisition);
static gint char_table_expose          (GtkWidget      *widget,
				      GdkEventExpose *event);
static gint char_table_button_press          (GtkWidget      *widget,
				      GdkEventButton *event);
static gint char_table_key_press	(GtkWidget      *widget,
				      GdkEventKey *event);
static void char_table_item_select(CharTable *table);
static void char_table_item_unselect(CharTable *table);
static void char_table_item_draw(CharTable *table, int x, int y, int selected);

static GtkWidgetClass *parent_class = NULL;

enum {
	SELECTED,
	LAST_SIGNAL
};

static gint char_table_signals[LAST_SIGNAL] = { 0 };

typedef void (*CharTableSignal1) (GtkObject *object,
	gpointer arg1,
	gpointer data);

static void char_table_marshal_signal_1 (GtkObject *object,
	GtkSignalFunc func,
	gpointer func_data,
	GtkArg	*arg);

guint
char_table_get_type ()
{
  static guint char_table_type = 0;

  if (!char_table_type)
    {
      GtkTypeInfo char_table_info =
      {
	"CharTable",
	sizeof (CharTable),
	sizeof (CharTableClass),
	(GtkClassInitFunc) char_table_class_init,
	(GtkObjectInitFunc) char_table_init,
	(GtkArgSetFunc) NULL,
	(GtkArgGetFunc) NULL,
      };

      char_table_type = gtk_type_unique (gtk_widget_get_type (), &char_table_info);
    }
  /* g_print("%s in %s\n", __FUNCTION__, __FILE__); */

  return char_table_type;
}

static void
char_table_class_init (CharTableClass *class)
{
	GtkWidgetClass *widget_class;
	GtkObjectClass *object_class;

	object_class = (GtkObjectClass*) class;
	widget_class = (GtkWidgetClass*) class;

	char_table_signals[SELECTED] = gtk_signal_new("selected",
		GTK_RUN_LAST,
		object_class->type,
		GTK_SIGNAL_OFFSET (CharTableClass, selected),
		char_table_marshal_signal_1, GTK_TYPE_NONE, 1,
		GTK_TYPE_POINTER);
	gtk_object_class_add_signals (object_class, char_table_signals, LAST_SIGNAL);
	parent_class = gtk_type_class (gtk_widget_get_type ());
	object_class->destroy = char_table_destroy;
	widget_class->realize = char_table_realize;
	widget_class->size_allocate = char_table_size_allocate;
	widget_class->size_request = char_table_size_request;
	widget_class->expose_event = char_table_expose;
	widget_class->button_press_event = char_table_button_press;
	widget_class->key_press_event = char_table_key_press;
}

static void
char_table_init (CharTable *chartable)
{
	GTK_WIDGET_SET_FLAGS(chartable, GTK_CAN_FOCUS);
	chartable->current_x = 0;
	chartable->current_y = 0;
}


GtkWidget *
char_table_new(char **buf, int col, int row)
{
	CharTable *table;

	g_return_val_if_fail(buf != NULL, NULL);
	g_return_val_if_fail(col > 0, NULL);
	g_return_val_if_fail(row > 0, NULL);

	table = gtk_type_new(char_table_get_type());
	table->col = col;
	table->row = row;
	table->buf = buf;
	return GTK_WIDGET(table);
}

char *
char_table_get_char(CharTable *table)
{
	g_return_val_if_fail(table != NULL, NULL);
	g_return_val_if_fail(IS_CHAR_TABLE(table), NULL);

	if(table->current_x < 0) return NULL;
	return &(table->buf[table->current_x][table->current_y*2]);
}

static void
char_table_realize (GtkWidget *widget)
{
	CharTable *table;
	GdkWindowAttr attributes;
	gint attributes_mask;
	GdkFont *font;
	g_return_if_fail (widget != NULL);

	table = CHAR_TABLE(widget);
	GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);

	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.visual = gtk_widget_get_visual (widget);
	attributes.colormap = gtk_widget_get_colormap (widget);
	attributes.event_mask = gtk_widget_get_events (widget);
	attributes.event_mask |= (GDK_EXPOSURE_MASK |
		GDK_BUTTON_PRESS_MASK |
		GDK_KEY_PRESS_MASK);

	attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

	widget->window = gdk_window_new (gtk_widget_get_parent_window(widget), &attributes, attributes_mask);
	gdk_window_set_user_data (widget->window, table);
	widget->style = gtk_style_attach (widget->style, widget->window);
	gdk_window_set_background(widget->window, &widget->style->white);
	font = GTK_WIDGET(table)->style->font;
	table->per_width = gdk_string_width(font, "") + PAD * 2;
	/*
	printf("per_width = %d\n", table->per_width);
	*/
	table->per_height = font->ascent + font->descent + PAD * 2;
	gtk_widget_queue_resize(widget);
}

static void
char_table_size_allocate (GtkWidget     *widget,
				GtkAllocation *allocation)
{
	g_return_if_fail (widget != NULL);
	widget->allocation = *allocation;

	if (GTK_WIDGET_REALIZED (widget)) {
		gdk_window_move_resize (widget->window,
			allocation->x, allocation->y,
			allocation->width, allocation->height);
    }
}

static gint
char_table_expose(GtkWidget *w, GdkEventExpose *ev)
{
	int x1, x2;
	int y1, y2;
	int i, j;
	CharTable *table;

	g_return_val_if_fail(w != NULL, FALSE);

	table = CHAR_TABLE(w);
	x1 = ev->area.x / table->per_width;
	x2 = (ev->area.x + ev->area.width) / table->per_width;
	y1 = ev->area.y / table->per_height;
	y2 = (ev->area.y + ev->area.height) / table->per_height;
	if(y2 >= table->row) y2 = table->row - 1;
	if(x2 >= table->col) x2 = table->col - 1;
	for(i=x1; i<= x2; i++) for(j=y1;j<=y2;j++) {
		if(i == table->current_x && j == table->current_y) {
			char_table_item_select(table);
		} else
			char_table_item_draw(table, i, j, FALSE);
	}
	return FALSE;
}

static gint
char_table_button_press(GtkWidget *w, GdkEventButton *ev)
{
	CharTable *table;
	int row, col;

	g_return_val_if_fail(w != NULL, FALSE);

	table = CHAR_TABLE(w);

	if (!GTK_WIDGET_HAS_FOCUS (w))
		gtk_widget_grab_focus (w);
	col = ev->x / (table->per_width);
	row = ev->y / (table->per_height);
	if(row < 0 || row >= table->row) return FALSE;
	if(col < 0 || col >= table->col) return FALSE;
	if(ev->type == GDK_2BUTTON_PRESS) {
		gtk_signal_emit(GTK_OBJECT(w), char_table_signals[SELECTED], 
			&table->buf[table->current_x][2*table->current_y]);
		return TRUE;
	}
	if(table->current_x != col || table->current_y != row) {
		char_table_item_unselect(table);
		table->current_x = col;
		table->current_y = row;
		char_table_item_select(table);
	}
	if(ev->type == GDK_2BUTTON_PRESS || ev->button == 2) {
		gtk_signal_emit(GTK_OBJECT(w), char_table_signals[SELECTED], 
			&table->buf[table->current_x][2*table->current_y]);
		return TRUE;
	}
	return TRUE;
}

static gint
char_table_key_press(GtkWidget *w, GdkEventKey *ev)
{
	CharTable *table;
	g_return_val_if_fail( w != NULL, FALSE);

	table = CHAR_TABLE(w);
	if(ev->keyval == GDK_Left) {
		char_table_item_unselect(table);
		if(table->current_x == 0) {
			table->current_x = table->col -1;
			if(table->current_y == 0) {
				table->current_y = table->row -1;
			} else 
				table->current_y--;
		} else
			table->current_x--;
	} else if(ev->keyval == GDK_Right) {
		char_table_item_unselect(table);
		if(table->current_x == table->col -1) {
			table->current_x = 0;
			if(table->current_y == (table->row -1)) {
				table->current_y = 0;
			} else
				table->current_y++;
		} else
			table->current_x++;
	} else if(ev->keyval == GDK_Up) {
		char_table_item_unselect(table);
		if(table->current_y == 0) {
			table->current_y = table->row -1;
		} else 
			table->current_y--;
	} else if(ev->keyval == GDK_Down) {
		char_table_item_unselect(table);
		if(table->current_y == (table->row -1)) {
			table->current_y = 0;
		} else
			table->current_y++;
	} else if(ev->keyval == GDK_space || ev->keyval == GDK_Return) {
		gtk_signal_emit(GTK_OBJECT(w), char_table_signals[SELECTED], 
			&table->buf[table->current_x][2*table->current_y]);
	} else
		return FALSE;
	char_table_item_select(table);
	return TRUE;
}

static void
char_table_destroy(GtkObject *object)
{
	g_return_if_fail(object != NULL);

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
char_table_size_request(GtkWidget *widget, GtkRequisition *requisition)
{
	CharTable *chartable;
	g_return_if_fail (widget != NULL);

	chartable = CHAR_TABLE(widget);
	/*g_print("%s in %s\n", __FUNCTION__, __FILE__); */
	requisition->width = chartable->col * chartable->per_width;
	requisition->height = chartable->row * chartable->per_height;
}

static void
char_table_item_select(CharTable *table)
{
	GtkWidget *w = GTK_WIDGET(table);
	if(table->current_x < 0 || table->current_y < 0) return;
	if(table->current_x >= table->col || table->current_y >= table->row) return;
	gdk_draw_rectangle(w->window, w->style->bg_gc[GTK_STATE_SELECTED],
		TRUE, 
		table->per_width * table->current_x + 1, /*  x */
		table->per_height * table->current_y + 1, /* y */
		table->per_width - 2,
		table->per_height - 2);
	char_table_item_draw(table, table->current_x, table->current_y, TRUE);
}

static void
char_table_item_unselect(CharTable *table)
{
	GtkWidget *w = GTK_WIDGET(table);
	if(table->current_x < 0 || table->current_y < 0) return;
	if(table->current_x >= table->col || table->current_y >= table->row) return;
	gdk_window_clear_area(w->window, 
		table->per_width * table->current_x,
		table->per_height * table->current_y,
		table->per_width, table->per_height);
	char_table_item_draw(table, table->current_x, table->current_y, FALSE);
}

static void
char_table_item_draw(CharTable *table, int x, int y, int selected)
{
	GtkWidget *w = GTK_WIDGET(table);
	gdk_draw_rectangle(w->window, w->style->black_gc,
		FALSE, 
		table->per_width * x, 
		table->per_height * y,
		table->per_width - 2,
		table->per_height - 2);
	if(selected) {
		gdk_draw_text(w->window, w->style->font, w->style->fg_gc[GTK_STATE_SELECTED],
			table->per_width * x + PAD,
			table->per_height * y + PAD + w->style->font->ascent,
			&table->buf[x][2*y], 2);
	} else {
		gdk_draw_text(w->window, w->style->font, w->style->fg_gc[GTK_STATE_NORMAL],
			table->per_width * x + PAD,
			table->per_height * y + PAD + w->style->font->ascent,
			&table->buf[x][2*y], 2);
	}
}

static void
char_table_marshal_signal_1 (GtkObject *object,
	GtkSignalFunc func,
	gpointer func_data,
	GtkArg *args)
{
	CharTableSignal1 rfunc;
	rfunc = (CharTableSignal1)func;
	(*rfunc) (object, GTK_VALUE_POINTER(args[0]), func_data);
}
